/*
 * Copyright (c) 1999 Tony Rogvall <tony@bluetail.com>
 *
 * Offscreen stuff using fblinear32 to store raw PIXELVAL's
 *
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "../device.h"
#include "genfont.h"
#include "fb.h"

/* specific vgalib driver entry points*/
static int  off_open(PSD psd);
static void off_close(PSD psd);
static void off_getscreeninfo(PSD psd,PSCREENINFO psi);
static void off_setpalette(PSD psd,int first,int count,RGBENTRY *pal);
static void off_drawpixel(PSD psd,COORD x, COORD y, PIXELVAL c);
static PIXELVAL off_readpixel(PSD psd,COORD x, COORD y);
static void off_drawhline(PSD psd,COORD x1, COORD x2, COORD y, PIXELVAL c);
static void off_drawvline(PSD psd,COORD x, COORD y1, COORD y2, PIXELVAL c);
static void off_fillrect(PSD psd,COORD x1,COORD y1,COORD x2,COORD y2,PIXELVAL c);
static void off_blit(PSD dstpsd,COORD destx,COORD desty,COORD w,COORD h,
		     PSD srcpsd,COORD srcx,COORD srcy,int op);

SCREENDEVICE offdev = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL,
    off_open,
    off_close,
    off_getscreeninfo,
    off_setpalette,
    off_drawpixel,
    off_readpixel,
    off_drawhline,
    off_drawvline,
    off_fillrect,
    gen_getfontinfo,
    gen_gettextsize,
    gen_gettextbits,
    off_blit,
    NULL
};

extern FBENTRY	fblinear1;
extern FBENTRY	fblinear2;
extern FBENTRY	fblinear4;
extern FBENTRY	fblinear8;
extern FBENTRY	fblinear16;
extern FBENTRY	fblinear32;

int drawing;  /* TEMPORARY hack */

static FBENTRY* off_drv;


/* ASSUME pds->xres & psd->yres are set */
static int off_open(PSD psd)
{
    PIXELVAL* ptr;

    if (psd->ncolors == 2)
	off_drv = &fblinear1;
    else if (psd->ncolors == 4)
	off_drv = &fblinear2;
    else if (psd->ncolors == 16)
	off_drv = &fblinear4;
    else if (psd->ncolors == 256)
	off_drv = &fblinear8;
    else if (psd->ncolors == 65536)
	off_drv = &fblinear16;
    else if (psd->ncolors > 65536)
	off_drv = &fblinear32;
    else {
	off_drv = NULL;
	return -1;
    }

    psd->linelen = (psd->xvirtres*psd->bpp+7)/8;
    /* allocate pixel memory */
    if ((ptr = (PIXELVAL *) malloc(psd->linelen * psd->yvirtres)) == NULL)
	return -1;
    psd->flags = PSF_MEMORY | PSF_HAVEBLIT;
    psd->addr = ptr;
    off_drv->init(psd);

    /* FIXME: we could path the fblin32 routines into PSD here
       and avoid an extra function call!!! */
    return 0;
}

static void off_close(PSD psd)
{
    free(psd->addr);
}

static void off_getscreeninfo(PSD psd, PSCREENINFO psi)
{
    psi->rows = psd->yvirtres;
    psi->cols = psd->xvirtres;
    psi->planes = psd->planes;
    psi->bpp = psd->bpp;
    psi->ncolors = psd->ncolors;
    psi->pixtype = psd->pixtype;
    psi->fonts = NUMBER_FONTS;
    psi->xdpcm = 0; /* does this work? */
    psi->ydpcm = 0; /* does this work? */
}

static void off_setpalette(PSD psd, int first, int count, RGBENTRY *pal)
{
    /* NOP */
}

static void off_drawpixel(PSD psd, COORD x, COORD y, PIXELVAL c)
{
    off_drv->drawpixel(psd, x, y, c);
}

static PIXELVAL off_readpixel(PSD psd, COORD x, COORD y)
{
    return off_drv->readpixel(psd, x, y);
}

static void off_drawhline(PSD psd,COORD x1, COORD x2, COORD y, PIXELVAL c)
{
    off_drv->drawhorzline(psd, x1, x2+1, y, c);
}

static void off_drawvline(PSD psd,COORD x, COORD y1, COORD y2, PIXELVAL c)
{
    off_drv->drawvertline(psd, x, y1, y2+1, c);
}

static void off_fillrect(PSD psd,COORD x1,COORD y1,
			 COORD x2,COORD y2,PIXELVAL c)
{
    COORD y;
    x2++;
    for (y = y1; y <= y2; y++) 
	off_drv->drawhorzline(psd, x1, x2, y, c);
}

static void off_blit(PSD dstpsd,COORD destx,COORD desty,COORD w,COORD h,
		     PSD srcpsd,COORD srcx,COORD srcy,int op)
{
    off_drv->blit(dstpsd, destx, desty, w, h, srcpsd, srcx, srcy, op);
}



