/*
 * This software is Copyright (C) 1998 by T.sqware - all rights limited
 * It is provided in to the public domain "as is", can be freely modified
 * as far as this copyight notice is kept unchanged, but does not imply
 * an endorsement by T.sqware of the product in which it is included.
 *
 *  $Id: uart.cv 1.2 1999/03/07 17:23:39 rps Exp rps $
 */

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>


#include <rtems.h>
#include <irq.h>
#include "pc_uart.h"

/*
 * Basic 16552 driver
 */

#define	PC_BaseBaud 	115200

    					/* enables interrupts */
/* #define	UART_PAUSE		pthread_yield  */
#define	UART_PAUSE


#define disable __asm__ __volatile__("cli")
#define enable __asm__ __volatile__("sti");
/*
#define inb(port)							\
({									\
        register int _inb_result;					\
									\
        asm volatile ("xorl %%eax,%%eax; inb %%dx,%%al" :		\
            "=a" (_inb_result) : "d" (port));				\
        _inb_result;							\
})

#define outb(port, data)						\
        asm volatile ("outb %%al,%%dx" : : "a" (data), "d" (port))


#define inw(port)							\
({									\
        register int _inbw_result;					\
									\
        asm volatile ("xorl %%eax,%%eax; inw %%dx,%%ax" :		\
            "=a" (_inbw_result) : "d" (port));				\
        _inbw_result;							\
})

#define outw(port, data)						\
        asm volatile ("outw %%ax,%%dx" : : "a" (data), "d" (port))
*/

#if 0
void pthread_yield()
{
  sleep( 1 );
  /* sched_yield();  */
}
#endif


/**********************************************************************************
 * 
 * DESCRIPTION: Routine to output a word as defined in FreeBSD.
 *
 * RETURNS: nothing.
 *
 **********************************************************************************/
void __inline m_outb( unsigned short io_addr, unsigned char out_data )
{
  outport_byte( io_addr, out_data );
}

/**********************************************************************************
 * 
 * DESCRIPTION: Routine to read a word as defined in FreeBSD.
 *
 * RETURNS: byte read.
 *
 **********************************************************************************/
static __inline unsigned char m_inb( unsigned short io_addr )
{
  unsigned short in_data;
  inport_byte( io_addr, in_data );
  return in_data;
}


extern rtems_unsigned32 volatile Clock_driver_ticks;
extern unsigned int clockbits( void );
/*
/////////////////////////////////////////////////////////////////////////////
// DESCRIPTION:  Returns time in raw clock counts, approx 838 ns per tick. 
/////////////////////////////////////////////////////////////////////////////
*/
u_long get_timer_ticks( void )
{
	u_long hi;
	u_long lo;
	do 
	{
     __asm__ ( "cli" );
	  hi = Clock_driver_ticks;
	  lo = clockbits();
     __asm__ ( "sti" );
	} while( hi != Clock_driver_ticks );
	return( hi << 16 ) - lo;
}



/* 
 * Macros to read/wirte register of uart, if configuration is
 * different just rewrite these macros
 */ 


static inline unsigned char uread(int base, unsigned int reg)
{
  register unsigned char val;
  val = m_inb( base+reg );
  return val;
}

static inline void uwrite(int base, int reg, unsigned int val)
{
  m_outb( base+reg, val );
}

void pc_clear_output_line(PC_UART *device, int line )
{
  unsigned int mcr;
  mcr = uread (device->base, MCR);
  /* make line go down */
  mcr &= ~line;
  uwrite(device->base, MCR, mcr );
}

void pc_set_output_line(PC_UART *device, int line )
{
  unsigned int mcr;
  mcr = uread (device->base, MCR);
  /* make line go up */
  mcr |= line;
  uwrite(device->base, MCR, mcr );
}


int pc_all_data_sent(PC_UART *device, int wait )
{
  int rc;
  do
  {
     rc = ( uread( device->base, LSR ) & TEMT ) && device->txDone;
	 if( rc )
    {
      break;
    }
  } while( wait );
/*
  if( device->txDone )
  {
    printf( "all sent rc=%d, %d, mcr=%x\n", 
    	rc, device->txDone , uread( device->base, LSR ) );
  }
*/
  return rc;
}


#ifdef UARTDEBUG
    static void
m_uartError(PC_UART *device)
{
  unsigned char uartStatus, dummy;

  uartStatus = uread(device->base, LSR);
  dummy = uread(device->base, RBR);

  device->error = uartStatus;
  if (uartStatus & OE)
    printk("********* Over run Error **********\n");
  if (uartStatus & PE)
    printk("********* Parity Error   **********\n");
  if (uartStatus & FE)
    printk("********* Framing Error  **********\n");
  if (uartStatus & BI)
    printk("********* Parity Error   **********\n");
  if (uartStatus & ERFIFO)
    printk("********* Error receive Fifo **********\n");

}
#else
inline void m_uartError(PC_UART *device)
{
  unsigned char uartStatus;
  uartStatus = uread(device->base, LSR);
  device->error = uartStatus;
  uartStatus = uread(device->base, RBR);
}
#endif


/* 
 * Reset all fields controling this serial port.
 * Nothing is perfomed at the hardware level.
 */
void pc_uart_flush( PC_UART *device )
{
  device->error = 0;
  device->modemStatus = 0;

  device->rxCount = 0;
  device->rxHead = device->rxBuf;
  device->rxTail = device->rxBuf;

  device->txCount = 0;
  device->txHead = device->txBuf;
  device->txTail = device->txBuf;
  device->xmiting = 0;
  device->txDone  = 1;
}

/* rtems_irq_connect_data irqInfo; */
/**********************************************************************************
 * 
 * DESCRIPTION:
 *
 * RETURNS:
 *
 **********************************************************************************/
static void nopOn(const rtems_irq_connect_data* notUsed)
{
  /* does nothing */
}

/**********************************************************************************
 * 
 * DESCRIPTION:
 *
 * RETURNS:
 *
 **********************************************************************************/
static int _IsOn(const rtems_irq_connect_data* irq)
{
  return BSP_irq_enabled_at_i8259s (irq->name);
}

/**********************************************************************************
 * 
 * DESCRIPTION: Register the interrupt handler with RTEMS.
 *
 * RETURNS:  0 - error; 1- success.
 *
 **********************************************************************************/
static int uart_initialize_irq( rtems_irq_hdl uart_interrupt_handler, int irq )
{
  rtems_status_code st;
  rtems_irq_connect_data irqInfo;

  irqInfo.hdl = uart_interrupt_handler;
  irqInfo.on  = nopOn;
  irqInfo.off = nopOn;
  irqInfo.isOn = _IsOn;
  irqInfo.name = irq;
 
  printf ("IRQ with Kernel: %d\n", irqInfo.name );
  st = ( rtems_status_code )BSP_install_rtems_irq_handler( &irqInfo );
  if( !st )
  {
    printf("Can't attach WD interrupt handler for irq %d\n", irqInfo.name );
    return 0;
  }
  return 1;
}


/* 
 * Uart initialization, it is hardcoded to 8 bit, no parity,
 * one stop bit, FIFO, things to be changed
 * are baud rate and nad hw flow control,
 * and longest rx fifo setting
 */
void pc_uart_init( PC_UART *data, rtems_irq_hdl irq_handler )
{
  unsigned char tmp;

  printf( "uart:init - baud=%d\n", data->baud );  
  switch(data->baud)
    {
    case 50:
    case 75:
    case 110:
    case 134:
    case 300:
    case 600:
    case 1200:
    case 2400:
    case 9600:
    case 19200:
    case 38400:
    case 57600:
    case 115200:
      break;
    default:
/*      assert(0);   */
      return;
    }


  /* Init Queues */
  data->rxEnd = data->rxBuf + data->rxSize;
  data->txEnd = data->txBuf + data->txSize;
  pc_uart_flush( data );

  
  /* Set DLAB bit to 1 */
  uwrite(data->base, LCR, DLAB);
  
  /* Set baud rate */
  uwrite(data->base, DLL,  (PC_BaseBaud/data->baud) & 0xff); 
  uwrite(data->base, DLM,  ((PC_BaseBaud/data->baud) >> 8) & 0xff); 

  /* 8-bit, no parity , 1 stop */
  uwrite( data->base, LCR, data->nchars );
  

  /* Set DTR, RTS and OUT2 high */
  uwrite( data->base, MCR, DTR | RTS | OUT_2 );

  /* Enable FIFO */
  uwrite(data->base, FCR, FIFO_EN | XMIT_RESET | RCV_RESET | RECEIVE_FIFO_TRIGGER12); 

  /* Disable Interrupts */
  uwrite(data->base, IER, 0);

  if( irq_handler )
  {
    uart_initialize_irq( irq_handler, data->irq );
  }
  /* Read status to clear them */
  tmp = uread(data->base, LSR);
  tmp = uread(data->base, RBR);

  data->modemStatus = uread(data->base, MSR);

  data->rx_thread_id = 0;
  data->tx_thread_id = 0;

  printf( "uart:init is complete\n" );  
  
  /* Remember state */
  return;
}

/* 
 * Set baud
 */
void pc_uart_set_baud( PC_UART *device, int baud )
{
  unsigned char mcr, ier;

  /* 
   * This function may be called whenever TERMIOS parameters
   * are changed, so we have to make sire that baud change is 
   * indeed required
   */

  if( device->baud == baud )
     return;


  mcr = uread(device->base, MCR);
  ier = uread(device->base, IER);
  
  device->baud = baud;

  /* Assumes the interrupt hanlder is already installed */
  pc_uart_init( device, 0 );

  uwrite( device->base, MCR, mcr );
  uwrite( device->base, IER, ier );
  
  return;
}

/*
 * Enable/disable interrupts 
 */
void pc_uart_intr_ctrl(PC_UART *device, int cmd)
{

  switch(cmd)
    {
    case BSP_UART_INTR_CTRL_DISABLE:
      uwrite(device->base, IER, INTERRUPT_DISABLE);
      break;

    case BSP_UART_INTR_CTRL_ENABLE:
      if( device->hwFlow )
	{
	  uwrite(device->base, IER,
		 (RECEIVE_ENABLE  |
		  TRANSMIT_ENABLE |
		  RECEIVER_LINE_ST_ENABLE |
		  MODEM_ENABLE
		 )
		);
	}
      else
	{
	  uwrite(device->base, IER,
		 (RECEIVE_ENABLE  |
		  TRANSMIT_ENABLE |
		  RECEIVER_LINE_ST_ENABLE
		 )
		);
	}
      break;
    case BSP_UART_INTR_CTRL_TERMIOS:
      if(device->hwFlow)
	{
	  uwrite(device->base, IER,
		 (RECEIVE_ENABLE  |
		  RECEIVER_LINE_ST_ENABLE |
		  MODEM_ENABLE
		 )
		);
	}
      else
	{
	  uwrite(device->base, IER,
		 (RECEIVE_ENABLE  |
		  RECEIVER_LINE_ST_ENABLE
		 )
		);
	}
      break;
    case BSP_UART_INTR_CTRL_GDB:
      uwrite(device->base, IER, RECEIVE_ENABLE);
      break;
    default:
/*      assert(0);   */
      break;
    }
 
  return;
}

void pc_uart_throttle(PC_UART *device)
{
  unsigned int mcr;
  
  if(!device->hwFlow)
    {
      /* Should not happen */
/*      assert(0);  */
      return;
    }
  mcr = uread (device->base, MCR);
  /* RTS down */
  mcr &= ~RTS;
  uwrite(device->base, MCR, mcr);

  return;
}

void pc_uart_unthrottle(PC_UART *device)
{
  unsigned int mcr;

  if(!device->hwFlow)
    {
      /* Should not happen */
/*      assert(0);   */
      return;
    }
  mcr = uread (device->base, MCR);
  /* RTS up */
  mcr |= RTS;
  uwrite(device->base, MCR, mcr);

  return;
}

/*
 * Status function, -1 if error
 * detected, 0 if no received chars available,
 * 1 if received char available, 2 if break
 * is detected, it will eat break and error 
 * chars. It ignores overruns - we cannot do 
 * anything about - it execpt count statistics
 * and we are not counting it.
 */
int pc_uart_polled_status(PC_UART *device)
{
  unsigned char val;

  val = uread(device->base, LSR);

  if(val & BI)
    {
      /* BREAK found, eat character */
      uread(device->base, RBR);
      return BSP_UART_STATUS_BREAK;
    }

  if((val & (DR | OE | FE)) ==  1)
    {
      /* No error, character present */ 
      return BSP_UART_STATUS_CHAR;
    }

  if((val & (DR | OE | FE)) == 0)
    {
      /* Nothing */
      return BSP_UART_STATUS_NOCHAR;
    }

  /* 
   * Framing or parity error eat character
   */
  uread( device->base, RBR );
 
  return BSP_UART_STATUS_ERROR;
}


/*
 * Polled mode write function
 */
void pc_uart_polled_write(PC_UART *device, int val)
{
  unsigned char val1;
  
  for(;;)
    {
      if((val1=uread(device->base, LSR)) & THRE)
	{
	  break;
	}
/*    UART_PAUSE();  */
    /*	printf( "/" ); */
    }

  if(device->hwFlow)
    {
      for(;;)
	{
	  if(uread(device->base, MSR) & CTS)
	    {
	      break;
	    }
	}
    }

  uwrite(device->base, THR, val & 0xff);
      
  return;
}

/* 
 * Polled mode read function
 */
int pc_uart_polled_read(PC_UART *device)
{
  unsigned char val;

  for(;;)
    {
      if(uread(device->base, LSR) & DR)
	{
	  break;
	}
    /*	printf( "-" );  */
/*	UART_PAUSE();  */
    }
  
  val = uread(device->base, RBR);

  return (int)(val & 0xff);
}


void pc_uart_enable_ier(int base, int iir )
{
   uwrite( base, IER, uread( base, IER ) | iir );
}

void pc_uart_disable_ier(int base, int iir )
{
   uwrite( base, IER, uread( base, IER ) & ( ~iir ) );
}

void pc_uart_isr(PC_UART *p)
{
  unsigned char val;
  int      vect;

  for(;;)
    {
      vect = uread(p->base, IIR) & 0xf;
      switch(vect)
	{
	case MODEM_STATUS :
	  val = uread( p->base, MSR );
      p->modemStatus = val;
	  break;

	case NO_MORE_INTR :
	  /* No more interrupts */
	  return;

	case TRANSMITTER_HODING_REGISTER_EMPTY :
	  /* 
	   * TX holding empty: we have to disable these interrupts 
	   * if there is nothing more to send. 
	   */
   	  /* If nothing else to send disable interrupts */
	  if( p->txCount == 0 )
      {
		  p->xmiting = 0;
		  p->txDone  = 1;
		  pc_uart_disable_ier( p->base, TRANSMIT_ENABLE );
       }
	   else
	   {
          uwrite( p->base, THR, *p->txHead & 0xff );
	      if( ++p->txHead > p->txEnd ) 
  	      	 p->txHead = p->txBuf;
	      p->txCount--;
	   }
      if( p->tx_thread_id )
      {
        rtems_event_send( p->tx_thread_id, PC_UART_XMIT_EVENT );
      }
	   break;

	case RECEIVER_DATA_AVAIL :
	case CHARACTER_TIMEOUT_INDICATION:
	  /* RX data ready */
     val = uread(p->base, RBR);
	  if( p->rxCount <= p->rxSize )
	  {
         *p->rxTail = val;
  	      if( ++p->rxTail > p->rxEnd ) 
  	          p->rxTail = p->rxBuf;
	      p->rxCount++;
        if( p->rx_thread_id )
        {
           rtems_event_send( p->rx_thread_id, PC_UART_RECEIVE_EVENT );
        }
	  }
	  break;

	case RECEIVER_ERROR:
	  /* RX error: eat character */
	   m_uartError( p );
	  break;
	default:
	  /* Should not happen */
/*	  assert(0); */
	  return;
	}
    }
}


int pc_uart_write_char( PC_UART *p, char ch )
{
  int ier;
  /* If there TX buffer is busy - something is royally screwed up */
  while( ( uread( p->base, LSR ) & THRE ) == 0 )
  		;

  /* CTS low */
  if( p->hwFlow && ( p->modemStatus & CTS ) )
	  return 0;

  ier = TRANSMIT_ENABLE;
  if(p->hwFlow )
     ier |= MODEM_ENABLE;
  p->xmiting = 1;
  p->txDone  = 0;
  pc_uart_enable_ier( p->base, ier );
  /* Write character */
  uwrite( p->base, THR, ch & 0xFF );
  return 1;
}



static int pc_uart_write_internal( PC_UART *p, char* buf, int size )
{
  int total;

  total = size;
  disable; 
  if( !p->xmiting )
  {
   pc_uart_write_char( p, *buf++ );
	size--;
  }
  while( size && ( p->txCount < p->txSize ) )
  {
    *p->txTail = *buf++;
    if( ++p->txTail > p->txEnd ) 
       p->txTail = p->txBuf;
    p->txCount++;
	size--;
  }
  enable;
  return total - size;
}

int pc_uart_write( PC_UART *p, char* buf, int size, u_long time_out )
{
  rtems_interval t1;
  rtems_interval elapsed;
  rtems_interval total_ticks;
  int  bytes_sent;
  
  /* if this is a no wait call ( time_out == 0 ) or there is free
     space in the tx buffer, go a head and queue up.
  */
  if( !time_out || ( ( p->txSize - p->txCount ) >= size ) )
      return pc_uart_write_internal( p, buf, size );

  /* Let's wait for more bytes, until they get available or times out... */
  p->tx_thread_id = RTEMS_SELF;
  total_ticks = TOD_MILLISECONDS_TO_TICKS( time_out );
  rtems_clock_get( RTEMS_CLOCK_GET_TICKS_SINCE_BOOT, ( rtems_interval *)&t1 );
  elapsed = 0;
  bytes_sent = 0;
  do
  {
     rtems_event_set o;
     rtems_status_code rc;
/*     printk( "TX Waiting: %d", total_ticks - elapsed );  */
     rc = rtems_event_receive( PC_UART_XMIT_EVENT, RTEMS_WAIT, total_ticks - elapsed, &o );
     if( ( ( p->txSize - p->txCount ) >= size ) || ( rc == RTEMS_TIMEOUT ) )
     {
        bytes_sent += pc_uart_write_internal( p, &buf[ bytes_sent], size - bytes_sent );
        if( size == bytes_sent )
            return bytes_sent;
           
     }
     /* if we got here, we have transmitted some bytes, but it is less than the
        caller asked for. So, let's recalculate the timeout that we have to
        wait...
     */
     rtems_clock_get( RTEMS_CLOCK_GET_TICKS_SINCE_BOOT, ( rtems_interval *)&elapsed );
     if( ( elapsed - t1 ) >= total_ticks )
     {
        bytes_sent += pc_uart_write_internal( p, &buf[ bytes_sent], size - bytes_sent );
        return bytes_sent;
     }
     elapsed = elapsed - t1;
  } while( 1 );
  return 0;
}



static int pc_uart_read_internal( PC_UART *p, char* buf, int size )
{
  int retSize;
  int i;
  retSize = p->rxCount;
  if( retSize > size )
      retSize = size;
  for( i=0; i < retSize; i++ )
  {
    *buf++ = *p->rxHead;
    disable; 
    if( ++p->rxHead > p->rxEnd ) 
       p->rxHead = p->rxBuf;
    p->rxCount--;
    enable;
  }
  return retSize;
}


int pc_uart_read( PC_UART *p, char* buf, int size, u_long time_out )
{
  rtems_interval t1;
  rtems_interval elapsed;
  rtems_interval total_ticks;
  
  /* if this is a no wait call ( time_out == 0 ) or there is more
     bytes in the buffer that the caller is asking for, return the
     bytes to the caller right away
  */
  if( !time_out || ( p->rxCount >= size ) )
     return pc_uart_read_internal( p, buf, size );

  /* Let's wait for more bytes, until they get available or times out... */
  p->rx_thread_id = _Thread_Executing->Object.id;
/* 
   printk( "RX task ID: %d", p->rx_thread_id );
   printk( "RX task ID ==>: %d", _Thread_Executing->Object.id );
 */
  total_ticks = TOD_MILLISECONDS_TO_TICKS( time_out );
  rtems_clock_get( RTEMS_CLOCK_GET_TICKS_SINCE_BOOT, ( rtems_interval *)&t1 );
  elapsed = 0;
  do
  {
     rtems_event_set o;
     rtems_status_code rc;
/*     printk( "RX Waiting: %d", total_ticks - elapsed );  */
     o = 0;
     rc = rtems_event_receive( PC_UART_RECEIVE_EVENT, RTEMS_WAIT, total_ticks - elapsed, &o );
/*     printk( "RX rc: %d", rc );  */
     if( ( p->rxCount >= size ) || ( rc == RTEMS_TIMEOUT ) )
     {
        return pc_uart_read_internal( p, buf, size );
     }
     /* if we got here, we have received some bytes, but it is less than the
        caller asked for. So, let's recalculate the timeout that we have to
        wait...
     */
/*     printk( "RX Enough bytes: %d ", p->rxCount );  */
     rtems_clock_get( RTEMS_CLOCK_GET_TICKS_SINCE_BOOT, ( rtems_interval *)&elapsed );
     if( ( elapsed - t1 ) > total_ticks )
     {
        return pc_uart_read_internal( p, buf, size );
     }
     elapsed = elapsed - t1;
/*     printk( "RX elapsed: %d", elapsed );  */
  } while( 1 );
  return 0;
}

	  



